"""
Paicos is an object-oriented Python package for analysis of (cosmological)
simulations performed with Arepo. Its main feature is that it includes
handling of units, derived variables, loading of data upon request.

The module includes a way of writing and reading data with units.
The code is parallel with an OpenMP Cython implementation.
"""

__version__ = "0.1.4"
__author__ = 'Thomas Berlok'
__credits__ = 'Leibniz-Institute for Astrophysics Potsdam (AIP)'

# Dependencies
import os
import numpy
import scipy
import h5py
import astropy

# Settings and utility functions
from . import util
from . import settings
from .util import root_dir

# HDF5 file readers
from .readers.arepo_snap import Snapshot
from .readers.arepo_catalog import Catalog
from .readers.paicos_readers import PaicosReader, ImageReader, Histogram2DReader

# HDF5 file writers
from .writers.paicos_writer import PaicosWriter, PaicosTimeSeriesWriter
from .writers.arepo_image import ArepoImage

# Image creators
from .image_creators.image_creator import ImageCreator
from .image_creators.projector import Projector
from .image_creators.nested_projector import NestedProjector
from .image_creators.tree_projector import TreeProjector
from .image_creators.slicer import Slicer

# Histograms
from .histograms.histogram import Histogram
from .histograms.histogram2D import Histogram2D

# Derived variables
from .derived_variables import derived_variables

# Cython functions
from . import cython

# pylint: disable=W0621


def use_units(use_units):
    """
    pa.use_units(True) turns on paicos quantities globally
    pa.use_units(False) loads in data without applying units.

    The status can be seen in pa.settings.use_units
    """
    settings.use_units = use_units


def add_user_function(variable_string, function):
    """
    This functions allows to enable user functions for obtaining
    deriving variables. An example could be the following:

    def TemperaturesTimesMassesSquared(snap, get_depencies=False):
        if get_depencies:
            return ['0_Temperatures', '0_Masses']
        return snap['0_Temperatures'] * snap['0_Masses']**2


    pa.add_user_function('0_TM2', TemperaturesTimesMassesSquared)

    """
    derived_variables.user_functions.update({variable_string: function})


def use_only_user_functions(use_only_user_functions):
    """
    Passing 'True' to this functions disables automatic derivation of
    variables except for the ones you have added via 'add_user_function'.
    """
    settings.use_only_user_functions = use_only_user_functions


def add_user_unit(field, blockname, unit):
    """
    This function adds user units.

    Parameters:

    field (string): possible values are specified in
                    unit_specifications.pos_fields, currently

                    pos_fields = ['default', 'voronoi_cells', 'dark_matter',
                    'stars', 'black_holes', 'groups', 'subhalos']

    blockname (string): The blockname that you would like to enable units for.

    unit (string): The unit that the blockname should have.
                                   String inputs are preferred, e.g.,

                         'arepo_mass arepo_length small_a^2 small_h^(-3/2)'

                   The available arepo units are

                   arepo_length (often 1 kpc)
                   arepo_mass (often 10^10 Msun)
                   arepo_velocity (often 1 km/s)
                   arepo_time
                   arepo_energy
                   arepo_pressure
                   arepo_density

                   Please note that these only become available once you
                   actually load a hdf5 file. This is the reason that you can
                   pass the string only.
    """
    if field not in util.user_unit_dict:
        raise RuntimeError(f'unknown field: {field}')

    util.user_unit_dict[field][blockname] = unit


def numthreads(numthreads):
    """
    Pass the number of threads you would like to use for parallel execution.

    numthreads (int): e.g. 16
    """
    util.check_if_omp_has_issues(False)

    if numthreads > settings.max_threads:
        print(f'Your machine only has {settings.max_threads} available threads')

    settings.numthreads = min(numthreads, settings.max_threads)
    if settings.openMP_has_issues:
        settings.numthreads_reduction = 1
    else:
        settings.numthreads_reduction = settings.numthreads


def print_info_when_deriving_variables(option):
    """
    Input: a boolean controlling whether to provide info to terminal.
    """
    settings.print_info_when_deriving_variables = option


def give_openMP_warnings(option):
    """
    Turns of warnings that will otherwise appear on a mac computer.
    """
    settings.give_openMP_warnings = option


def set_aliases(aliases):
    """
    Assign a list of aliases for use as keys in snapshot objects.

    input (dictionary): e.g.

    aliases = {'0_Density': 'dens',
           '0_Temperatures': 'T',
           '0_MeanMolecularWeight': 'mu'}
    """
    msg = 'Your user-set aliases seem to not be unique'
    assert len(set(aliases.values())) == len(aliases.values()), msg

    inverse = {aliases[key]: key for key in aliases.keys()}
    settings.aliases = aliases
    settings.inverse_aliases = inverse
    settings.use_aliases = True


def user_settings_exists():
    """
    Checks if user settings exist in the root directory of Paicos.
    """
    if os.path.exists(root_dir + 'user_settings.py'):
        return True
    return False


if user_settings_exists():
    # pylint: disable=E0401
    import user_settings

# Do this at start up
util.check_if_omp_has_issues()



"""
This defines a base class for image creators (such as Projector and Slicer)
"""

import numpy as np
from .. import util
from .. import settings
from .. import units


class ImageCreator:
    """
    This is a base class for creating images from a snapshot.
    """

    def __init__(self, snap, center, widths, direction, npix=512, parttype=0):
        """
        Initialize the ImageCreator class. This method will be called
        by subclasses such as Projector or Slicer.

        Parameters:
            snap (object): Snapshot object from which the image is created

            center: Center of the image (3D coordinates)

            widths: Width of the image in each direction (3D coordinates)

            direction (str): Direction of the image ('x', 'y', 'z')

            npix (int): Number of pixels in the image (default is 512)
        """

        self.snap = snap

        code_length = self.snap.length

        if hasattr(center, 'unit'):
            self.center = center
            assert center.unit == code_length.unit
        elif settings.use_units:
            self.center = np.array(center) * code_length
        else:
            self.center = np.array(center)

        if hasattr(widths, 'unit'):
            self.widths = widths
            assert widths.unit == code_length.unit
        elif settings.use_units:
            self.widths = np.array(widths) * code_length
        else:
            self.widths = np.array(widths)

        self.x_c = self.center[0]
        self.y_c = self.center[1]
        self.z_c = self.center[2]
        self.width_x = self.widths[0]
        self.width_y = self.widths[1]
        self.width_z = self.widths[2]

        self.direction = direction

        self.npix = self.npix_width = npix

        self.parttype = parttype

        if direction == 'x':
            extent = [self.y_c - self.width_y / 2, self.y_c + self.width_y / 2,
                      self.z_c - self.width_z / 2, self.z_c + self.width_z / 2]

            centered_extent = [- self.width_y / 2, self.width_y / 2,
                               - self.width_z / 2, self.width_z / 2]

            self.width = self.width_y
            self.height = self.width_z
            self.depth = self.width_x

        elif direction == 'y':
            extent = [self.x_c - self.width_x / 2, self.x_c + self.width_x / 2,
                      self.z_c - self.width_z / 2, self.z_c + self.width_z / 2]

            centered_extent = [- self.width_x / 2, self.width_x / 2,
                               - self.width_z / 2, self.width_z / 2]

            self.width = self.width_x
            self.height = self.width_z
            self.depth = self.width_y

        elif direction == 'z':
            extent = [self.x_c - self.width_x / 2, self.x_c + self.width_x / 2,
                      self.y_c - self.width_y / 2, self.y_c + self.width_y / 2]

            centered_extent = [- self.width_x / 2, self.width_x / 2,
                               - self.width_y / 2, self.width_y / 2]

            self.width = self.width_x
            self.height = self.width_y
            self.depth = self.width_z

        if settings.use_units:
            a = snap._Time
            h = snap.h
            comoving = snap.comoving_sim
            self.extent = units.PaicosQuantity(extent, a=a, h=h, comoving_sim=comoving)

            self.centered_extent = units.PaicosQuantity(centered_extent, a=a, h=h,
                                                        comoving_sim=comoving)
            self.npix_height = int((self.height / self.width).value * self.npix_width)
        else:
            self.extent = np.array(extent)
            self.centered_extent = np.array(centered_extent)
            self.npix_height = int((self.height / self.width) * self.npix_width)

        area = (self.extent[1] - self.extent[0]) * (self.extent[3] - self.extent[2])
        self.area = area
        self.area_per_pixel = self.area / (self.npix_width * self.npix_height)
        self.volume = self.width_x * self.width_y * self.width_z
        self.volume_per_pixel = self.volume / (self.npix_width * self.npix_height)



"""
Defines a class that creates an image of a given variable by projecting it
onto a 2D plane.
"""
import numpy as np
from .image_creator import ImageCreator
from .. import util
from .. import settings
from .. import units


class Projector(ImageCreator):
    """
    A class that allows creating an image of a given variable by projecting
    it onto a 2D plane.

    The Projector class is a subclass of the ImageCreator class.
    The Projector class creates an image of a given variable by projecting
    it onto a 2D plane.

    It takes in several parameters such as a snapshot object, center and
    widths of the region, direction of projection, and various optional
    parameters for number of pixels, smoothing length and number of threads
    for parallelization. It then calls various functions from the paicos
    package, including get_index_of_region, load_data, and get_volumes. This
    class also has a function called _check_if_omp_has_issues which checks if
    the parallelization via OpenMP works, and sets the number of threads
    accordingly.
    """

    def __init__(self, snap, center, widths, direction,
                 npix=512, parttype=0, nvol=8, make_snap_with_selection=True):
        """
        Initialize the Projector class.

        Parameters
        ----------
        snap : Snapshot
            A snapshot object of Snapshot class from paicos package.

        center : numpy array
            Center of the region on which projection is to be done, e.g.
            center = [x_c, y_c, z_c].

        widths : numpy array
            Widths of the region on which projection is to be done,
            e.g.m widths=[width_x, width_y, width_z].

        direction : str
            Direction of the projection, e.g. 'x', 'y' or 'z'.

        npix : int, optional
            Number of pixels in the horizontal direction of the image,
            by default 512.

        parttype : int, optional
            Number of the particle type to project, by default gas (PartType 0).

        nvol : int, optional
            Integer used to determine the smoothing length, by default 8

        """

        # call the superclass constructor to initialize the ImageCreator class
        super().__init__(snap, center, widths, direction, npix=npix, parttype=parttype)

        parttype = self.parttype

        # nvol is an integer that determines the smoothing length
        self.nvol = nvol

        # get the index of the region of projection
        self.index = util.get_index_of_cubic_region(self.snap[f"{parttype}_Coordinates"],
                                                    center, widths, snap.box)

        # Reduce the snapshot to only contain region of interest
        if make_snap_with_selection:
            self.snap = self.snap.select(self.index, parttype=parttype)

        # Calculate the smoothing length
        avail_list = (list(snap.keys()) + snap._auto_list)
        if f'{parttype}_Volume' in avail_list:
            self.hsml = np.cbrt(nvol * (self.snap[f"{parttype}_Volume"])
                                / (4.0 * np.pi / 3.0))
        elif f'{parttype}_SubfindHsml' in avail_list:
            self.hsml = self.snap[f'{parttype}_SubfindHsml']
        else:
            raise RuntimeError('There is no smoothing length or volume for the projector')

        self.pos = self.snap[f'{self.parttype}_Coordinates']

        if not make_snap_with_selection:
            self.hsml = self.hsml[self.index]
            self.pos = self.pos[self.index]

    @util.remove_astro_units
    def _cython_project(self, center, widths, variable):
        """
        Private method for projecting using cython
        """
        if settings.openMP_has_issues:
            from ..cython.sph_projectors import project_image as project
        else:
            from ..cython.sph_projectors import project_image_omp as project

        x_c, y_c, z_c = center[0], center[1], center[2]
        width_x, width_y, width_z = widths

        boxsize = self.snap.box
        if self.direction == 'x':
            projection = project(self.pos[:, 1],
                                 self.pos[:, 2],
                                 variable,
                                 self.hsml, self.npix,
                                 y_c, z_c, width_y, width_z,
                                 boxsize, settings.numthreads_reduction)
        elif self.direction == 'y':
            projection = project(self.pos[:, 0],
                                 self.pos[:, 2],
                                 variable,
                                 self.hsml, self.npix,
                                 x_c, z_c, width_x, width_z,
                                 boxsize, settings.numthreads_reduction)
        elif self.direction == 'z':
            projection = project(self.pos[:, 0],
                                 self.pos[:, 1],
                                 variable,
                                 self.hsml, self.npix,
                                 x_c, y_c, width_x, width_y,
                                 boxsize, settings.numthreads_reduction)

        return projection

    def project_variable(self, variable):
        """
        projects a given variable onto a 2D plane.

        Parameters
        ----------
        variable : str, function, numpy array
            variable, it can be passed as string or an array

        Returns
        -------
        numpy array
            The image of the projected variable
        """

        if isinstance(variable, str):
            err_msg = 'projector uses a different parttype'
            assert int(variable[0]) == self.parttype, err_msg
            variable = self.snap[variable]
        else:
            if not isinstance(variable, np.ndarray):
                raise RuntimeError('Unexpected type for variable')

        if variable.shape == self.index.shape:
            variable = variable[self.index]
        # Do the projection
        projection = self._cython_project(self.center, self.widths, variable)

        # Transpose
        projection = projection.T

        assert projection.shape[0] == self.npix_height
        assert projection.shape[1] == self.npix_width

        area_per_pixel = self.area / np.prod(projection.shape)

        if isinstance(variable, units.PaicosQuantity):
            projection = projection * variable.unit_quantity

        return projection / area_per_pixel
